@@ -2,27 +2,29 @@
 /*                                                                          */
 /*  The FreeType project -- a free and portable quality TrueType renderer.  */
 /*                                                                          */
-/*  Copyright (C) 1996-2020 by                                              */
+/*  Copyright (C) 1996-2024 by                                              */
 /*  D. Turner, R.Wilhelm, and W. Lemberg                                    */
 /*                                                                          */
+/*                                                                          */
+/*  ftdump: a dumper of basic font information                              */
+/*                                                                          */
 /****************************************************************************/
 
 
 #include <ft2build.h>
-#include FT_FREETYPE_H
-#include FT_SFNT_NAMES_H
-#include FT_TRUETYPE_IDS_H
-#include FT_TRUETYPE_TABLES_H
-#include FT_TRUETYPE_TAGS_H
-#include FT_MULTIPLE_MASTERS_H
+#include <freetype/freetype.h>
 
-  /* the following header shouldn't be used in normal programs */
-#include <freetype/internal/ftdebug.h>
+#include <freetype/ftbdf.h>
+#include <freetype/ftmm.h>
+#include <freetype/ftmodapi.h>  /* showing driver name */
+#include <freetype/ftsnames.h>
+#include <freetype/ftwinfnt.h>
+#include <freetype/ttnameid.h>
+#include <freetype/tttables.h>
+#include <freetype/tttags.h>
+#include <freetype/t1tables.h>
 
-  /* showing driver name */
-#include FT_MODULE_H
-#include <freetype/internal/ftobjs.h>
-#include <freetype/internal/ftdrv.h>
+#include <freetype/ftcid.h>
 
   /* error messages */
 #undef FTERRORS_H_
@@ -61,7 +63,7 @@
     FT_Done_FreeType( library );
 
     switch( error )
-    #include FT_ERRORS_H
+    #include <freetype/fterrors.h>
 
     fprintf( stderr, "%s\n  error = 0x%04x, %s\n", message, error, str );
     exit( 1 );
@@ -80,23 +82,41 @@
 
 
   static void
-  usage( FT_Library  library,
-         char*       execname )
+  Print_Array( FT_Short*  data,
+               FT_Byte    num_data )
+  {
+    FT_Int  i;
+
+
+    printf( "[" );
+    if ( num_data )
+    {
+      printf( "%d", data[0] );
+      for ( i = 1; i < num_data; i++ )
+        printf( ", %d", data[i] );
+    }
+    printf( "]\n" );
+  }
+
+
+  static void
+  usage( FT_Library   library,
+         const char*  execname )
   {
     FT_Done_FreeType( library );
 
     fprintf( stderr,
       "\n"
       "ftdump: simple font dumper -- part of the FreeType project\n"
-      "-----------------------------------------------------------\n"
+      "----------------------------------------------------------\n"
       "\n"
       "Usage: %s [options] fontname\n"
       "\n",
              execname );
 
     fprintf( stderr,
-      "  -c, -C    Print charmap coverage.\n"
-      "  -n        Print SFNT name tables.\n"
+      "  -c, -C    Print charmap coverage and/or CID coverage.\n"
+      "  -n        Print SFNT 'name' table or Type1 font info.\n"
       "  -p        Print TrueType programs.\n"
       "  -t        Print SFNT table list.\n"
       "  -u        Emit UTF8.\n"
@@ -152,8 +172,8 @@
     if ( head )
     {
       char    buf[11];
-      time_t  created  = head->Created [1];
-      time_t  modified = head->Modified[1];
+      time_t  created  = (time_t)head->Created [1];
+      time_t  modified = (time_t)head->Modified[1];
 
 
       /* ignore most of upper bits until 2176 and adjust epoch */
@@ -162,13 +182,20 @@
       modified = head->Modified[0] == 1 ? modified + 2212122496U
                                         : modified - 2082844800U;
 
-      strftime( buf, sizeof ( buf ), "%Y-%m-%d", gmtime( &created  ) );
-      printf( "%s%s\n", Name_Field( "created" ), buf );
-      strftime( buf, sizeof ( buf ), "%Y-%m-%d", gmtime( &modified ) );
-      printf( "%s%s\n", Name_Field( "modified" ), buf );
+      /* ignore pre-epoch time that gmtime cannot handle on some systems */
+      if ( created >= 0 )
+      {
+        strftime( buf, sizeof ( buf ), "%Y-%m-%d", gmtime( &created ) );
+        printf( "%s%s\n", Name_Field( "created" ), buf );
+      }
+      if ( modified >= 0 )
+      {
+        strftime( buf, sizeof ( buf ), "%Y-%m-%d", gmtime( &modified ) );
+        printf( "%s%s\n", Name_Field( "modified" ), buf );
+      }
 
-      printf( "%s%.2f\n", Name_Field( "revision" ),
-              head->Font_Revision / 65536.0 );
+      printf( head->Font_Revision & 0xFFC0 ? "%s%.4g\n" : "%s%.2f\n",
+              Name_Field( "revision" ), head->Font_Revision / 65536.0 );
     }
   }
 
@@ -176,14 +203,10 @@
   static void
   Print_Type( FT_Face  face )
   {
-    FT_Module  module;
-
-
     printf( "font type entries\n" );
 
-    module = &face->driver->root;
     printf( "%s%s\n", Name_Field( "FreeType driver" ),
-            module->clazz->module_name );
+            FT_FACE_DRIVER_NAME( face ) );
 
     /* Is it better to dump all sfnt tag names? */
     printf( "%s%s\n", Name_Field( "sfnt wrapped" ),
@@ -432,6 +455,214 @@
   }
 
 
+  static void
+  Print_FontInfo_Dictionary( PS_FontInfo  fi )
+  {
+    printf( "/FontInfo dictionary\n" );
+
+    printf( "%s%s\n", Name_Field( "FamilyName" ),
+            fi->family_name );
+    printf( "%s%s\n", Name_Field( "FullName" ),
+            fi->full_name );
+    printf( "%s%d\n", Name_Field( "isFixedPitch" ),
+            fi->is_fixed_pitch );
+    printf( "%s%ld\n", Name_Field( "ItalicAngle" ),
+            fi->italic_angle );
+    printf( "%s%s\n", Name_Field( "Notice" ),
+            fi->notice );
+    printf( "%s%d\n", Name_Field( "UnderlinePosition" ),
+            fi->underline_position );
+    printf( "%s%u\n", Name_Field( "UnderlineThickness" ),
+            fi->underline_thickness );
+    printf( "%s%s\n", Name_Field( "version" ),
+            fi->version );
+    printf( "%s%s\n", Name_Field( "Weight" ),
+            fi->weight );
+  }
+
+
+  /*
+   * FreeType 2 API supports 32-bit gid, but
+   * the CIDFont does not support 32-bit CID,
+   * because of the 64k limit of the array
+   * and dictionary objects in PostScript.
+   */
+#ifndef FT_CID_MAX
+#define FT_CID_MAX 0xFFFFU
+#endif
+  /*
+   * Print a range specified by 2 integers.
+   */
+  static void
+  Print_UInt_Range( FT_UInt  from,
+                    FT_UInt  to,
+                    char*    is_first )
+  {
+    if (!(*is_first))
+      printf(",");
+
+    if ( from == to )
+      printf( "%d", from );
+    else if ( from < to )
+      printf( "%d-%d", from, to );
+
+    *is_first = 0;
+  }
+
+
+  /*
+   * Print implemented CIDs by calling
+   *   FT_Get_CID_From_Glyph_Index() for all GIDs.
+   *
+   */
+  static void
+  Print_CIDs( FT_Face  face )
+  {
+    FT_UInt  gid = 0, max_gid = FT_UINT_MAX;
+    FT_UInt  cid = 0, rng_from = 0, rng_to = 0;
+    char     is_first_rng = 1;
+    
+
+    if ( face->num_glyphs < 1 )
+      return;
+
+    printf( "\n" );
+    printf( "CID coverage\n" );
+    printf( "   " );
+
+    if ( (FT_ULong)face->num_glyphs < FT_UINT_MAX )
+      max_gid = (FT_UInt)face->num_glyphs;
+
+    for ( gid = 0; gid <= max_gid; gid ++ )
+    {
+      if ( FT_Get_CID_From_Glyph_Index( face, gid, &cid ) )
+        continue;
+
+      if ( FT_CID_MAX < cid )
+      {
+        fprintf( stderr, "gid=%d resulted too large CID=%d, ignore it\n", gid, cid );
+        break;
+      }
+
+      if ( rng_to == cid )
+        continue;
+      else if ( cid < rng_to )
+      {
+        fprintf( stderr, "Unordered GID-CID map is found, please file your issue on "
+                         "https://gitlab.freedesktop.org/groups/freetype/-/issues\n" );
+        exit( 1 );
+      }
+      else if ( rng_to + 1 == cid )
+      {
+        rng_to = cid;
+        continue;
+      }
+
+      /* Found a gap (rng_to + 1 < cid), print the last range */
+      Print_UInt_Range( rng_from, rng_to, &is_first_rng );
+      rng_to = rng_from = cid;
+    }
+
+    Print_UInt_Range( rng_from, rng_to, &is_first_rng );
+
+    printf( "\n" );
+  }
+
+
+  /*
+   * Print_CIDFontInfo_Dictionary() might be conventional,
+   * but other tables, like gcid, can have ROS info too.
+   */
+  static void
+  Print_ROS_From_Face( FT_Face  face )
+  {
+    FT_Bool      is_cid = 0;
+    const char*  r = NULL;
+    const char*  o = NULL;
+    FT_Int       s = -1;
+
+
+    if ( FT_Get_CID_Is_Internally_CID_Keyed( face, &is_cid ) )
+      return;
+
+    if ( !is_cid )
+      return;
+
+    if ( FT_Get_CID_Registry_Ordering_Supplement( face, &r, &o, &s ) )
+      return;
+
+    printf( "\n" );
+    printf( "/CIDSystemInfo dictionary\n" );
+
+    if ( r )
+      printf( "%s%s\n", Name_Field( "Registry" ), r );
+
+    if ( o )
+      printf( "%s%s\n", Name_Field( "Ordering" ), o );
+
+    printf( "%s%d\n", Name_Field( "Supplement" ), s );
+
+    if ( coverage > 0 )
+      Print_CIDs( face );
+  }
+
+
+  static void
+  Print_FontPrivate_Dictionary( PS_Private  fp )
+  {
+    printf( "/Private dictionary\n" );
+
+    printf( "%s%d\n", Name_Field( "BlueFuzz" ),
+            fp->blue_fuzz );
+    printf( "%s%.6f\n", Name_Field( "BlueScale" ),
+            (double)fp->blue_scale / 65536 / 1000 );
+    printf( "%s%d\n", Name_Field( "BlueShift" ),
+            fp->blue_shift );
+    printf( "%s", Name_Field( "BlueValues" ) );
+    Print_Array( fp->blue_values,
+                 fp->num_blue_values );
+    printf( "%s%.4f\n", Name_Field( "ExpansionFactor" ),
+            (double)fp->expansion_factor / 65536 );
+    printf( "%s", Name_Field( "FamilyBlues" ) );
+    Print_Array( fp->family_blues,
+                 fp->num_family_blues );
+    printf( "%s", Name_Field( "FamilyOtherBlues" ) );
+    Print_Array( fp->family_other_blues,
+                 fp->num_family_other_blues );
+    printf( "%s%s\n", Name_Field( "ForceBold" ),
+            fp->force_bold ? "true" : "false" );
+    printf( "%s%ld\n", Name_Field( "LanguageGroup" ),
+            fp->language_group );
+    printf( "%s%d\n", Name_Field( "lenIV" ),
+            fp->lenIV );
+    printf( "%s", Name_Field( "MinFeature" ) );
+    Print_Array( fp->min_feature,
+                 2 );
+    printf( "%s", Name_Field( "OtherBlues" ) );
+    Print_Array( fp->other_blues,
+                 fp->num_other_blues );
+    printf( "%s%ld\n", Name_Field( "password" ),
+            fp->password );
+    printf( "%s%s\n", Name_Field( "RndStemUp" ),
+            fp->round_stem_up ? "true" : "false" );
+    /* casting to `FT_Short` is not really correct, but... */
+    printf( "%s", Name_Field( "StdHW" ) );
+    Print_Array( (FT_Short*)fp->standard_width,
+                 1 );
+    printf( "%s", Name_Field( "StdVW" ) );
+    Print_Array( (FT_Short*)fp->standard_height,
+                 1 );
+    printf( "%s", Name_Field( "StemSnapH" ) );
+    Print_Array( fp->snap_widths,
+                 fp->num_snap_widths );
+    printf( "%s", Name_Field( "StemSnapV" ) );
+    Print_Array( fp->snap_heights,
+                 fp->num_snap_heights );
+    printf( "%s%d\n", Name_Field( "UniqueID" ),
+            fp->unique_id );
+  }
+
+
   static void
   Print_Sfnt_Tables( FT_Face  face )
   {
@@ -456,7 +687,7 @@
       else
         continue;
 
-      printf( "  %2lu: %c%c%c%c %02X%02X%02X%02X...\n", i,
+      printf( "  %2lu: %c%c%c%c   %02X %02X %02X %02X ...\n", i,
                                    (FT_Char)( tag >> 24 ),
                                    (FT_Char)( tag >> 16 ),
                                    (FT_Char)( tag >>  8 ),
@@ -507,33 +738,44 @@
 
     for( i = 0; i < face->num_charmaps; i++ )
     {
-      FT_Long   format  = FT_Get_CMap_Format( face->charmaps[i] );
-      FT_ULong  lang_id = FT_Get_CMap_Language_ID( face->charmaps[i] );
+      FT_CharMap   cmap = face->charmaps[i];
+      FT_Long      format  = FT_Get_CMap_Format( cmap );
+      FT_ULong     lang_id = FT_Get_CMap_Language_ID( cmap );
+      const char*  encoding;
+      const char*  registry;
 
+      FT_WinFNT_HeaderRec  header;
+
+
+      printf( cmap->encoding ? " %c%2d: %c%c%c%c"
+                             : " %c%2d: none",
+              i == active ? '*' : ' ',
+              i,
+              cmap->encoding >> 24,
+              cmap->encoding >> 16,
+              cmap->encoding >> 8,
+              cmap->encoding );
+
+      printf( ", platform %u, encoding %2u",
+              cmap->platform_id,
+              cmap->encoding_id );
 
       if ( format >= 0 )
-        printf( "  %2d: format %2ld, platform %u, encoding %2u",
-                i,
-                format,
-                face->charmaps[i]->platform_id,
-                face->charmaps[i]->encoding_id );
-      else
-        printf( "  %2d: synthetic, platform %u, encoding %2u",
-                i,
-                face->charmaps[i]->platform_id,
-                face->charmaps[i]->encoding_id );
-
-      if ( lang_id == 0xFFFFFFFFUL )
-        printf( "   (Unicode Variation Sequences)" );
-      else
-        printf( "   language %lu",
-                lang_id );
-
-      if ( i == active )
-        printf( " (active)" );
+        printf( lang_id != 0xFFFFFFFFUL ? ", format %2lu, language %lu "
+                                        : ", format %2lu, UVS",
+                format, lang_id );
+      else if ( !FT_Get_BDF_Charset_ID( face, &encoding, &registry ) )
+        printf( ", charset %s-%s", registry, encoding );
+      else if ( !FT_Get_WinFNT_Header( face, &header ) )
+        printf( header.charset < 10 ? ", charset %hhu"
+                                    : ", charset %hhu <%hhX>",
+                header.charset, header.charset );
 
       printf ( "\n" );
 
+      if ( lang_id == 0xFFFFFFFFUL )  /* nothing further for UVS */
+        continue;
+
       if ( coverage == 2 )
       {
         FT_ULong   charcode;
@@ -541,7 +783,7 @@
         FT_String  buf[32];
 
 
-        FT_Set_Charmap( face, face->charmaps[i] );
+        FT_Set_Charmap( face, cmap );
 
         charcode = FT_Get_First_Char( face, &gindex );
         while ( gindex )
@@ -558,7 +800,7 @@
       }
       else if ( coverage == 1 )
       {
-        FT_ULong  next, last = ~1;
+        FT_ULong  next, last = ~1U;
         FT_UInt   gindex;
 
         const char*  f1 = "";
@@ -566,7 +808,7 @@
         const char*  f3 = "";
 
 
-        FT_Set_Charmap( face, face->charmaps[i] );
+        FT_Set_Charmap( face, cmap );
 
         next = FT_Get_First_Char( face, &gindex );
         while ( gindex )
@@ -596,19 +838,53 @@
 
 
   static void
-  Print_MM_Axes( FT_Face  face )
+  get_english_name_entry( FT_Face       face,
+                          FT_UInt       strid,
+                          FT_SfntName*  entry )
+  {
+    FT_UInt  num_names = FT_Get_Sfnt_Name_Count( face );
+    FT_UInt  i;
+
+    FT_SfntName  name = { 0, 0, 0, 0, NULL, 0 };
+
+
+    for ( i = 0; i < num_names; i++ )
+    {
+      error = FT_Get_Sfnt_Name( face, i, &name );
+      if ( error )
+        continue;
+
+      if ( name.name_id == strid )
+      {
+        /* XXX we don't have support for Apple's new `ltag' table yet, */
+        /* thus we ignore TT_PLATFORM_APPLE_UNICODE                    */
+        if ( ( name.platform_id == TT_PLATFORM_MACINTOSH &&
+               name.language_id == TT_MAC_LANGID_ENGLISH )        ||
+             ( name.platform_id == TT_PLATFORM_MICROSOFT        &&
+               ( name.language_id & 0xFF )
+                                == TT_MS_LANGID_ENGLISH_GENERAL ) )
+          break;
+      }
+    }
+
+    if ( i < num_names )
+      *entry = name;
+  }
+
+
+  static void
+  Print_MM_Info( FT_Face  face )
   {
     FT_MM_Var*       mm;
     FT_Multi_Master  dummy;
-    FT_UInt          is_GX, i, num_names;
+    FT_SfntName      name;
+    FT_UInt          is_GX, i;
 
 
     /* MM or GX axes */
     error = FT_Get_Multi_Master( face, &dummy );
     is_GX = error ? 1 : 0;
 
-    printf( "%s axes\n", is_GX ? "GX" : "MM" );
-
     error = FT_Get_MM_Var( face, &mm );
     if ( error )
     {
@@ -616,60 +892,134 @@
       return;
     }
 
-    num_names = FT_Get_Sfnt_Name_Count( face );
+    printf( "%s info\n", is_GX ? "GX" : "MM" );
+
+    printf( "  axes (%u)\n", mm->num_axis );
 
     for ( i = 0; i < mm->num_axis; i++ )
     {
-      FT_SfntName  name;
-
-
       name.string = NULL;
 
       if ( is_GX )
-      {
-        FT_UInt  strid = mm->axis[i].strid;
-        FT_UInt  j;
-
-
-        /* iterate over all name entries        */
-        /* to find an English entry for `strid' */
-
-        for ( j = 0; j < num_names; j++ )
-        {
-          error = FT_Get_Sfnt_Name( face, j, &name );
-          if ( error )
-            continue;
-
-          if ( name.name_id == strid )
-          {
-            /* XXX we don't have support for Apple's new `ltag' table yet, */
-            /* thus we ignore TT_PLATFORM_APPLE_UNICODE                    */
-            if ( ( name.platform_id == TT_PLATFORM_MACINTOSH &&
-                   name.language_id == TT_MAC_LANGID_ENGLISH )        ||
-                 ( name.platform_id == TT_PLATFORM_MICROSOFT        &&
-                   ( name.language_id & 0xFF )
-                                    == TT_MS_LANGID_ENGLISH_GENERAL ) )
-              break;
-          }
-        }
-      }
+        get_english_name_entry( face, mm->axis[i].strid, &name );
 
+      printf( "    %u: ", i );
       if ( name.string )
       {
         if ( name.platform_id == TT_PLATFORM_MACINTOSH )
-          put_ascii( name.string, name.string_len, 3 );
+          put_ascii( name.string, name.string_len, 0 );
         else
-          put_unicode_be16( name.string, name.string_len, 3, utf8 );
+          put_unicode_be16( name.string, name.string_len, 0, utf8 );
       }
       else
-        printf( "   %s", mm->axis[i].name );
+        printf( "%s", mm->axis[i].name );
 
-      printf( ": [%g;%g], default %g\n",
+      printf( ", [%g;%g], default %g\n",
               mm->axis[i].minimum / 65536.0,
               mm->axis[i].maximum / 65536.0,
               mm->axis[i].def / 65536.0 );
     }
 
+    if ( is_GX )
+    {
+      FT_Fixed*    coords;
+      const char*  ps_name;
+
+      FT_UInt              instance_count;
+      FT_UInt              default_named_instance;
+      FT_Var_Named_Style*  named_styles;
+
+
+      /* Show Variation PostScript Name Prefix. */
+
+      coords = (FT_Fixed*)malloc( mm->num_axis * sizeof ( FT_Fixed ) );
+      if ( coords == NULL )
+        return;
+
+      /* We temporarily activate variation font handling.  Because we */
+      /* use the default axes, the now retrieved PS name is identical */
+      /* to the PS name prefix.                                       */
+      FT_Get_Var_Design_Coordinates( face, mm->num_axis, coords );
+      FT_Set_Var_Design_Coordinates( face, mm->num_axis, coords );
+
+      ps_name = FT_Get_Postscript_Name( face );
+      if ( ps_name == NULL )
+        ps_name = "UNAVAILABLE";
+
+      printf( "\n"
+              "  VF PS name prefix: %s\n", ps_name );
+
+      /* Switch off variation font handling. */
+      FT_Set_Var_Design_Coordinates( face, 0, NULL );
+
+      free( coords );
+
+
+      /* Show named instances. */
+
+      instance_count = (FT_UInt)face->style_flags >> 16;
+      named_styles   = mm->namedstyle;
+
+      FT_Get_Default_Named_Instance( face, &default_named_instance );
+      default_named_instance--;   /* `named_styles` is a zero-based array */
+
+      printf( "\n" );
+      printf( "  named instances (%u)\n", instance_count );
+
+      for ( i = 0; i < instance_count; i++ )
+      {
+        int        pos;
+        FT_UInt    j;
+        FT_Bool    semicolon;
+        FT_Fixed*  c;
+
+
+        /* Since FreeType starts the instance numbering with value 1 */
+        /* in `face_index` we report the same here for consistency.  */
+        pos = printf( "    %u: ", i + 1);
+
+        name.string = NULL;
+        get_english_name_entry( face, named_styles[i].strid, &name );
+        if ( name.string )
+        {
+          if ( name.platform_id == TT_PLATFORM_MACINTOSH )
+            put_ascii( name.string, name.string_len, 0 );
+          else
+            put_unicode_be16( name.string, name.string_len, 0, utf8 );
+        }
+        else
+          printf( "UNAVAILABLE" );
+        printf( "%s\n", i == default_named_instance ? " (default)" : "" );
+
+        name.string = NULL;
+        get_english_name_entry( face, named_styles[i].psid, &name );
+        printf( "%*s   PS: ", pos, "" );
+        if ( name.string )
+        {
+          if ( name.platform_id == TT_PLATFORM_MACINTOSH )
+            put_ascii( name.string, name.string_len, 0 );
+          else
+            put_unicode_be16( name.string, name.string_len, 0, utf8 );
+        }
+        else
+          printf( "UNAVAILABLE" );
+        printf( "\n" );
+
+        semicolon = 0;
+        c         = named_styles[i].coords;
+
+        printf( "%*scoord: (", pos, "" );
+        for ( j = 0; j < mm->num_axis; j++ )
+        {
+          printf( "%s%g", semicolon ? ";" : "", c[j] / 65536.0);
+          semicolon = 1;
+        }
+        printf( ")\n" );
+      }
+
+      printf( "\n" );
+    }
+
     FT_Done_MM_Var( face->glyph->library, mm );
   }
 
@@ -722,7 +1072,7 @@
   {
     FT_ULong    fpgm_length = 0;
     FT_ULong    prep_length = 0;
-    FT_ULong    loca_length = 0;
+    FT_ULong    loca_length;
     FT_ULong    glyf_length = 0;
     FT_UShort   i;
     FT_Byte*    buffer = NULL;
@@ -764,6 +1114,23 @@
     Print_Bytecode( buffer, (FT_UShort)prep_length, "prep" );
 
   Glyf:
+    head =     (TT_Header*)FT_Get_Sfnt_Table( face, FT_SFNT_HEAD );
+    maxp = (TT_MaxProfile*)FT_Get_Sfnt_Table( face, FT_SFNT_MAXP );
+
+    if ( head == NULL || maxp == NULL )
+      goto Exit;
+
+    loca_length = head->Index_To_Loc_Format ? 4 * maxp->numGlyphs + 4
+                                            : 2 * maxp->numGlyphs + 2;
+
+    offset = (FT_Byte*)malloc( loca_length );
+    if ( offset == NULL )
+      goto Exit;
+
+    error = FT_Load_Sfnt_Table( face, TTAG_loca, 0, offset, &loca_length );
+    if ( error )
+      goto Exit;
+
     error = FT_Load_Sfnt_Table( face, TTAG_glyf, 0, NULL, &glyf_length );
     if ( error || glyf_length == 0 )
       goto Exit;
@@ -776,40 +1143,41 @@
     if ( error )
       goto Exit;
 
-    error = FT_Load_Sfnt_Table( face, TTAG_loca, 0, NULL, &loca_length );
-    if ( error || loca_length == 0 )
-      goto Exit;
-
-    offset = (FT_Byte*)malloc( loca_length );
-    if ( offset == NULL )
-      goto Exit;
-
-    error = FT_Load_Sfnt_Table( face, TTAG_loca, 0, offset, &loca_length );
-    if ( error )
-      goto Exit;
-
-    head =     (TT_Header*)FT_Get_Sfnt_Table( face, FT_SFNT_HEAD );
-    maxp = (TT_MaxProfile*)FT_Get_Sfnt_Table( face, FT_SFNT_MAXP );
-
     for ( i = 0; i < maxp->numGlyphs; i++ )
     {
-      FT_UInt32  loc;
+      FT_UInt32  loc, end;
       FT_UInt16  len;
       char       tag[5];
 
 
       if ( head->Index_To_Loc_Format )
+      {
         loc = (FT_UInt32)offset[4 * i    ] << 24 |
               (FT_UInt32)offset[4 * i + 1] << 16 |
               (FT_UInt32)offset[4 * i + 2] << 8  |
               (FT_UInt32)offset[4 * i + 3];
+        end = (FT_UInt32)offset[4 * i + 4] << 24 |
+              (FT_UInt32)offset[4 * i + 5] << 16 |
+              (FT_UInt32)offset[4 * i + 6] << 8  |
+              (FT_UInt32)offset[4 * i + 7];
+      }
       else
+      {
         loc = (FT_UInt32)offset[2 * i    ] << 9 |
               (FT_UInt32)offset[2 * i + 1] << 1;
+        end = (FT_UInt32)offset[2 * i + 2] << 9 |
+              (FT_UInt32)offset[2 * i + 3] << 1;
+      }
 
-      if ( loc >= glyf_length )
+      if ( end > glyf_length )
+        end = glyf_length;
+
+      if ( loc == end )
+        continue;
+
+      if ( loc + 1 >= end )
       {
-        printf( "\nglyf program %hd: invalid offset (%d)\n", i, loc );
+        printf( "\nglyph %hd: invalid offset (%d)\n", i, loc );
         continue;
       }
 
@@ -823,9 +1191,9 @@
 
         do
         {
-          if ( loc >= glyf_length )
+          if ( loc + 1 >= end )
           {
-            printf( "\nglyf program %hd: invalid offset (%d)\n", i, loc );
+            printf( "\nglyph %hd: invalid offset (%d)\n", i, loc );
             goto Continue;
           }
 
@@ -846,9 +1214,11 @@
       else
         loc += 2 * len;
 
-      if ( loc >= glyf_length )
+      if ( loc + 1 >= end )
       {
-        printf( "\nglyf program %hd: invalid offset (%d)\n", i, loc );
+        /* zero-contour glyphs can have no data */
+        if ( len )
+          printf( "\nglyph %hd: invalid offset (%d)\n", i, loc );
         continue;
       }
 
@@ -859,14 +1229,14 @@
 
       loc += 2;
 
-      if ( len >= glyf_length || loc >= glyf_length - len )
+      if ( loc + len > end )
       {
-        printf( "\nglyf program %hd: invalid size (%d)\n", i, len );
+        printf( "\nglyph %hd: invalid size (%d)\n", i, len );
         continue;
       }
 
       snprintf( tag, sizeof ( tag ), "%04hx", i );
-      printf( "\nglyf program %hd (%.4s)", i, tag );
+      printf( "\nglyph %hd (%.4s)", i, tag );
       Print_Bytecode( buffer + loc, len, tag );
 
     Continue:
@@ -882,7 +1252,7 @@
   static void
   Print_Glyfs( FT_Face  face )
   {
-    FT_ULong    loca_length = 0;
+    FT_ULong    loca_length;
     FT_ULong    glyf_length = 0;
     FT_UShort   i;
     FT_Byte*    buffer = NULL;
@@ -892,11 +1262,29 @@
     FT_Int      simple_overlap    = 0;
     FT_Int      composite         = 0;
     FT_Int      composite_overlap = 0;
+    FT_Int      empty             = 0;
 
     TT_Header*      head;
     TT_MaxProfile*  maxp;
 
 
+    head =     (TT_Header*)FT_Get_Sfnt_Table( face, FT_SFNT_HEAD );
+    maxp = (TT_MaxProfile*)FT_Get_Sfnt_Table( face, FT_SFNT_MAXP );
+
+    if ( head == NULL || maxp == NULL )
+      return;
+
+    loca_length = head->Index_To_Loc_Format ? 4 * maxp->numGlyphs + 4
+                                            : 2 * maxp->numGlyphs + 2;
+
+    offset = (FT_Byte*)malloc( loca_length );
+    if ( offset == NULL )
+      goto Exit;
+
+    error = FT_Load_Sfnt_Table( face, TTAG_loca, 0, offset, &loca_length );
+    if ( error )
+      goto Exit;
+
     error = FT_Load_Sfnt_Table( face, TTAG_glyf, 0, NULL, &glyf_length );
     if ( error || glyf_length == 0 )
       goto Exit;
@@ -909,40 +1297,44 @@
     if ( error )
       goto Exit;
 
-    error = FT_Load_Sfnt_Table( face, TTAG_loca, 0, NULL, &loca_length );
-    if ( error || loca_length == 0 )
-      goto Exit;
-
-    offset = (FT_Byte*)malloc( loca_length );
-    if ( offset == NULL )
-      goto Exit;
-
-    error = FT_Load_Sfnt_Table( face, TTAG_loca, 0, offset, &loca_length );
-    if ( error )
-      goto Exit;
-
-    head =     (TT_Header*)FT_Get_Sfnt_Table( face, FT_SFNT_HEAD );
-    maxp = (TT_MaxProfile*)FT_Get_Sfnt_Table( face, FT_SFNT_MAXP );
-
     for ( i = 0; i < maxp->numGlyphs; i++ )
     {
-      FT_UInt32  loc;
+      FT_UInt32  loc, end;
       FT_UInt16  len;
       FT_UShort  flags;
 
 
       if ( head->Index_To_Loc_Format )
+      {
         loc = (FT_UInt32)offset[4 * i    ] << 24 |
               (FT_UInt32)offset[4 * i + 1] << 16 |
               (FT_UInt32)offset[4 * i + 2] << 8  |
               (FT_UInt32)offset[4 * i + 3];
+        end = (FT_UInt32)offset[4 * i + 4] << 24 |
+              (FT_UInt32)offset[4 * i + 5] << 16 |
+              (FT_UInt32)offset[4 * i + 6] << 8  |
+              (FT_UInt32)offset[4 * i + 7];
+      }
       else
+      {
         loc = (FT_UInt32)offset[2 * i    ] << 9 |
               (FT_UInt32)offset[2 * i + 1] << 1;
+        end = (FT_UInt32)offset[2 * i + 2] << 9 |
+              (FT_UInt32)offset[2 * i + 3] << 1;
+      }
 
-      if ( loc >= glyf_length )
+      if ( end > glyf_length )
+        end = glyf_length;
+
+      if ( loc == end )
       {
-        printf( "\nglyf program %hd: invalid offset (%d)\n", i, loc );
+        empty++;
+        continue;
+      }
+
+      if ( end == 0 || loc >= end - 1 )
+      {
+        printf( "\nglyph %hd: invalid offset (%u)\n", i, loc );
         continue;
       }
 
@@ -953,9 +1345,9 @@
       {
         composite++;
 
-        if ( loc >= glyf_length )
+        if ( loc + 1 >= end )
         {
-          printf( "\nglyf program %hd: invalid offset (%d)\n", i, loc );
+          printf( "\nglyph %hd: invalid offset (%u)\n", i, loc );
           continue;
         }
 
@@ -970,9 +1362,11 @@
 
       loc += 2 * len;
 
-      if ( loc >= glyf_length )
+      if ( loc + 1 >= end )
       {
-        printf( "\nglyf program %hd: invalid offset (%d)\n", i, loc );
+        /* zero-contour glyphs can have no data */
+        if ( len )
+          printf( "\nglyph %hd: invalid offset (%d)\n", i, loc );
         continue;
       }
 
@@ -980,9 +1374,9 @@
 
       loc += 2 + len;
 
-      if ( len >= glyf_length )
+      if ( len >= end )
       {
-        printf( "\nglyf program %hd: invalid offset (%d)\n", i, loc );
+        printf( "\nglyph %hd: invalid offset (%d)\n", i, loc );
         continue;
       }
 
@@ -999,6 +1393,8 @@
     printf( composite_overlap ? ", with overlap flagged in %d\n"
                               : "\n",
             composite_overlap );
+    if ( empty )
+      printf( "%s%d\n", Name_Field( "   empty" ), empty );
 
   Exit:
     free( buffer );
@@ -1012,15 +1408,14 @@
   {
     int    i, file;
     char   filename[1024];
-    char*  execname;
     int    num_faces;
     int    option;
 
     FT_Library  library;      /* the FreeType library */
     FT_Face     face;         /* the font face        */
 
+    const char*  execname = ft_basename( argv[0] );
 
-    execname = ft_basename( argv[0] );
 
     /* Initialize engine */
     error = FT_Init_FreeType( &library );
@@ -1121,10 +1516,9 @@
     num_faces = face->num_faces;
     FT_Done_Face( face );
 
-    printf( "There %s %d %s in this file.\n",
-            num_faces == 1 ? (char *)"is" : (char *)"are",
-            num_faces,
-            num_faces == 1 ? (char *)"face" : (char *)"faces" );
+    printf( num_faces == 1 ? "There is %d face in %s.\n"
+                           : "There are %d faces in %s.\n",
+            num_faces, ft_basename( filename ) );
 
     for ( i = 0; i < num_faces; i++ )
     {
@@ -1142,10 +1536,29 @@
       printf( "\n" );
       Print_Type( face );
 
-      if ( name_tables && FT_IS_SFNT( face ) )
+      if ( name_tables )
       {
-        printf( "\n" );
-        Print_Sfnt_Names( face );
+        PS_FontInfoRec  font_info;
+        PS_PrivateRec   font_private;
+
+
+        if ( FT_IS_SFNT( face ) )
+        {
+          printf( "\n" );
+          Print_Sfnt_Names( face );
+        }
+
+        if ( FT_Get_PS_Font_Info( face, &font_info ) == FT_Err_Ok )
+        {
+          printf( "\n" );
+          Print_FontInfo_Dictionary( &font_info );
+        }
+
+        if ( FT_Get_PS_Font_Private( face, &font_private ) == FT_Err_Ok )
+        {
+          printf( "\n" );
+          Print_FontPrivate_Dictionary( &font_private );
+        }
       }
 
       if ( tables && FT_IS_SFNT( face ) )
@@ -1172,10 +1585,15 @@
         Print_Charmaps( face );
       }
 
+      /* FT_IS_CID_KEYED() does not catch an OpenType/CFF,
+       * let Print_ROS_From_Face() catch various cases.
+       */
+      Print_ROS_From_Face( face );
+
       if ( FT_HAS_MULTIPLE_MASTERS( face ) )
       {
         printf( "\n" );
-        Print_MM_Axes( face );
+        Print_MM_Info( face );
       }
 
       FT_Done_Face( face );
